pub mod prime_utils {
    use num::{BigUint, One};
    use once_cell::sync::Lazy;

    #[must_use]
    pub fn log_2(x: &BigUint) -> u64 {
        x.bits() - 1
    }

    static FOUR: Lazy<BigUint> = Lazy::new(|| BigUint::from(4u8));
    static TWO: Lazy<BigUint> = Lazy::new(|| BigUint::from(2u8));
    static ONE: Lazy<BigUint> = Lazy::new(BigUint::one);
    const ZERO: BigUint = BigUint::ZERO;

    #[must_use]
    pub fn is_prime(number: &BigUint) -> bool {
        if &*ONE >= number {
            return false;
        }
        if &*FOUR > number {
            return true;
        }
        if !number.bit(0) {
            return false;
        }

        // mersenne numbers = 2^a - 1
        // a = log2(number + 1)
        let a = log_2(&(number + 1u8));

        let is_mersenne = {
            let num_p2 = {
                let mut num = BigUint::default();
                num.set_bit(a, true);
                num
            };
            assert!(num_p2 > BigUint::one());

            let num_p2 = num_p2 - &*ONE;

            &num_p2 == number
        };

        if is_mersenne {
            return mersenne_check(number, a);
        }

        let sqrtnum = number.sqrt() + &*ONE; //fake ceil function

        inner_prime(number, &sqrtnum)
    }

    fn mersenne_check(number: &BigUint, a: u64) -> bool {
        let mut last = FOUR.clone();

        for _ in 2..a {
            last = (last.pow(2) - &*TWO) % number;
        }

        last == ZERO
    }

    fn inner_prime(number: &BigUint, sqrtnum: &BigUint) -> bool {
        let mut i = BigUint::one();

        loop {
            i += &*TWO; //we begin checking 3 and then continue with every second number as every even number is divisible by 2 and we already checked that
            if &i >= sqrtnum {
                return true;
            }
            if number % &i == ZERO {
                return false;
            }
        }
    }
}
